import os

import six
import argparse

import ovation.upload as upload
import ovation.download as download
import ovation.contents as contents
import ovation.session as session


def main():
    parser = argparse.ArgumentParser(prog='python -m ovation.cli')
    parser.add_argument('-u', '--user', help='Ovation user email')

    subparsers = parser.add_subparsers(title='Available subcommands',
                                       description='Use `python -m ovation.cli <subcommand> -h` for additional help')

    # UPLOAD
    parser_upload = subparsers.add_parser('upload', description='Upload files to Ovation')
    parser_upload.add_argument('parent_id', help='Project or Folder UUID')
    parser_upload.add_argument('paths', nargs='+', help='Paths to local files or directories')
    parser_upload.set_defaults(func=upload.upload_paths)
    # upload.upload_paths(user=args.user,
    #                     parent_id=args.parent_id,
    #                     paths=args.paths)


    # DOWNLOAD
    parser_download = subparsers.add_parser('download', description='Download files from Ovation')
    parser_download.add_argument('entity_id', help='File or Revision UUID')
    parser_download.add_argument('-o', '--output', help='Output directory')
    parser_download.set_defaults(func=download.download_main)
    # download.download_main(user=args.user,
    #               entity_id=args.entity_id,
    #               output=args.output)


    # CONTENTS
    parser_ls = subparsers.add_parser('list',
                                      aliases=['ls'],
                                      description='List Projects or Contents')

    parser_ls.add_argument('parent_id', nargs='?', help='(Optional) Project or Folder ID', default='')

    parser_ls.set_defaults(func=contents.list_contents_main)

    args = parser.parse_args()

    if args.user is None:
        args.user = input('Email: ')

    if 'OVATION_PASSWORD' in os.environ:
        password = os.environ['OVATION_PASSWORD']
    else:
        password = None

    s = session.connect(args.user, password=password)
    args.session = s

    args.func(args)


if __name__ == '__main__':
    main()



import mimetypes
import os.path
import threading

import boto3
import six
import argparse
import os

import ovation.core as core

from ovation.session import connect
from tqdm import tqdm


class ProgressPercentage(object):
    def __init__(self, filename, progress=tqdm):
        self._filename = filename
        self._seen_so_far = 0
        self._lock = threading.Lock()
        self._size = float(os.path.getsize(filename))
        self._progress = progress(unit='B',
                                  unit_scale=True,
                                  total=self._size,
                                  desc=os.path.basename(filename))

    def __call__(self, bytes_amount):
        # To simplify we'll assume this is hooked up
        # to a single filename.
        with self._lock:
            self._seen_so_far += bytes_amount
            self._progress.update(self._seen_so_far)
            if self._seen_so_far >= self._size:
                self._progress.close()


def upload_folder(session, parent, directory_path, progress=tqdm):
    """
    Recursively uploads a folder to Ovation

    :param session: Session
    :param parent: Project or Folder root
    :param directory_path: local path to directory
    :param progress: if not None, wrap in a progress (i.e. tqdm). Default: tqdm
    """

    root_folder = parent
    for root, dirs, files in os.walk(directory_path):
        root_name = os.path.basename(root)
        if len(root_name) == 0:
            root_name = os.path.basename(os.path.dirname(root))

        root_folder = core.create_folder(session, root_folder, root_name)

        for f in files:
            path = os.path.join(root, f)
            file = core.create_file(session, root_folder, f)

            upload_revision(session, file, path, progress=progress)

def upload_file(session, parent, file_path, progress=tqdm):
    """
    Upload a file to Ovation

    :param session: Session
    :param parent: Project or Folder root
    :param file_path: local path to file
    :param progress: if not None, wrap in a progress (i.e. tqdm). Default: tqdm
    :return: created File entity dictionary
    """
    name = os.path.basename(file_path)
    file = core.create_file(session, parent, name)
    return upload_revision(session, file, file_path, progress=progress)


def guess_content_type(file_name):
    content_type = mimetypes.guess_type(file_name)[0]
    if content_type is None:
        content_type = 'application/octet-stream'

    return content_type

def upload_revision(session, parent_file, local_path, progress=tqdm):
    """
    Upload a new `Revision` to `parent_file`. File is uploaded from `local_path` to
    the Ovation cloud, and the newly created `Revision` version is set.
    :param session: ovation.connection.Session
    :param parent_file: file entity dictionary or file ID string
    :param local_path: local path
    :param progress: if not None, wrap in a progress (i.e. tqdm). Default: tqdm
    :return: new `Revision` entity dicitonary
    """

    if isinstance(parent_file, six.string_types):
        parent_file = session.get(session.entity_path('file', entity_id=parent_file))

    file_name = os.path.basename(local_path)
    content_type = guess_content_type(file_name)

    r = session.post(parent_file['links']['self'],
                     data={'entities': [{'type': 'Revision',
                                         'attributes': {'name': file_name,
                                                        'content_type': content_type}}]})
    revision = r['entities'][0]
    aws = r['aws'][0]['aws']

    aws_session = boto3.Session(aws_access_key_id=aws['access_key_id'],
                                aws_secret_access_key=aws['secret_access_key'],
                                aws_session_token=aws['session_token'])
    s3 = aws_session.resource('s3')

    file_obj = s3.Object(aws['bucket'], aws['key'])

    args = {'ContentType': content_type,
            'ServerSideEncryption': 'AES256'}
    if progress and os.path.exists(local_path):
        file_obj.upload_file(local_path, ExtraArgs=args,
                             Callback=ProgressPercentage(local_path, progress=progress))
    else:
        file_obj.upload_file(local_path, ExtraArgs=args)

    return session.put(revision['links']['upload-complete'], entity=None)



def upload_paths(args):
    session = args.session
    parent_id = args.parent_id
    paths = args.paths
    if paths is None:
        return

    for p in paths:
        print('Uploading {}'.format(p))
        if os.path.isdir(p):
            upload_folder(session, parent_id, p)
        else:
            upload_file(session, parent_id, p)




import os.path

import functools
import requests
import six

import ovation.core as core

from tqdm import tqdm
from six.moves.urllib_parse import urlsplit
from pprint import pprint
from multiprocessing.pool import ThreadPool as Pool


def revision_download_info(session, revision):
    """
    Get temporary download link and ETag for a Revision.

    :param session: ovation.connection.Session
    :param revision: revision entity dictionary or revision ID string
    :return: dict with `url`, `etag`, and S3 `path`
    """

    if isinstance(revision, six.string_types):
        e = session.get(session.entity_path('entities', entity_id=revision))
        if e.type == core.REVISION_TYPE:
            revision = e
        elif e.type == core.FILE_TYPE:
            revision = session.get(e.links.heads)[0]
        else:
            raise Exception("Whoops! {} is not a File or Revision".format(revision))

    if revision['type'] == core.FILE_TYPE:
        revision = session.get(revision.links.heads)[0]

    if not revision['type'] == core.REVISION_TYPE:
        raise Exception("Whoops! {} is not a File or Revision".format(revision['_id']))

    r = session.session.get(revision['attributes']['url'],
                            headers={'accept': 'application/json'},
                            params={'token': session.token})
    r.raise_for_status()

    return r.json()


def download_revision(session, revision, output=None, progress=tqdm):
    """
    Downloads a Revision to the local file system. If output is provided, file is downloaded
    to the output path. Otherwise, it is downloaded to the current working directory.

    If a File (entity or ID) is provided, the HEAD revision is downloaded.

    :param session: ovation.connection.Session
    :param revision: revision entity dictionary or ID string, or file entity dictionary or ID string
    :param output: path to folder to save downloaded revision
    :param progress: if not None, wrap in a progress (i.e. tqdm). Default: tqdm
    :return: file path
    """

    url = revision_download_info(session, revision)['url']
    response = requests.get(url, stream=True)

    name = os.path.basename(urlsplit(url).path)
    if output:
        dest = os.path.join(output, name)
    else:
        dest = name

    with open(dest, "wb") as f:
        if progress:
            for data in progress(response.iter_content(),
                                 unit='B',
                                 unit_scale=True,
                                 miniters=1):
                f.write(data)
        else:
            for data in response.iter_content():
                f.write(data)


def _download_revision_path(session, revision_path, progress=tqdm):
    return download_revision(session, revision_path[1], output=revision_path[0], progress=progress)


def download_folder(session, folder, output=None, progress=tqdm):
    files = _traverse_folder(session, folder, output=output, progress=progress)
    with Pool() as p:
        for f in progress(p.imap_unordered(functools.partial(_download_revision_path,
                                                             session,
                                                             progress=None),
                                           files),
                          desc='Downloading files',
                          unit='file',
                          total=len(files)):
            pass

def _traverse_folder(session, folder, output=None, progress=tqdm):
    folder = core.get_entity(session, folder)
    if folder is None:
        return

    # make folder
    if output is None:
        path = folder.attributes.name
    else:
        path = os.path.join(output, folder.attributes.name)

    if not os.path.isdir(path):
        os.mkdir(path)

    # get files
    files = [[path, f] for f in session.get(folder.relationships.files.related)]

    # for each folders, recurse and return files
    folders = session.get(folder.relationships.folders.related)
    with Pool() as p:
        for subfiles in progress(p.imap_unordered(functools.partial(_traverse_folder,
                                                                    session,
                                                                    output=path,
                                                                    progress=progress),
                                                  folders),
                                 desc='Traversing folders',
                                 unit='folder',
                                 total=len(folders)):
            files += subfiles

    return files


def download_main(args):
    session = args.session
    entity_id = args.entity_id
    output = args.output

    entity = core.get_entity(session, entity_id)
    if entity.type == core.FOLDER_TYPE or entity.type == core.PROJECT_TYPE:
        download_folder(session, entity, output=output)
    else:
        download_revision(session, entity_id, output=output)



import argparse
import json
import os.path
import boto3
import logging
import csv

import functools

import ovation.core as core
import ovation.upload as upload

from ovation.session import connect
from ovation.session import Session
from tqdm import tqdm
from multiprocessing import Manager
from multiprocessing.pool import ThreadPool as Pool

# Checkpoing CSV columns
PATH = 'path'
ID = 'id'
FIELDNAMES = [PATH, ID]

# Multiprocessing pool size
POOL_SIZE = int(os.environ['TRANSFER_POOL_SIZE']) if 'TRANSFER_POOL_SIZE' in os.environ else 5


def copy_bucket_contents(session, project=None, aws_access_key_id=None, aws_secret_access_key=None,
                         source_s3_bucket=None, destination_s3_bucket=None, progress=tqdm,
                         copy_file_fn=None, checkpoint=None):
    """

    :param session:
    :param project:
    :param aws_access_key_id:
    :param aws_secret_access_key:
    :param source_s3_bucket:
    :param destination_s3_bucket:
    :param progress: show process progress (i.e. tqdm). Default: tqdm
    :return: {'folders': [ created folders ], 'files': [ copied files ] }
    """

    src_s3_session = boto3.Session(aws_access_key_id=aws_access_key_id,
                                   aws_secret_access_key=aws_secret_access_key)

    src_s3_connection = src_s3_session.resource('s3')
    src = src_s3_connection.Bucket(source_s3_bucket)

    logging.info('Starting copy from s3 bucket: ' + str(source_s3_bucket))

    logging.info('Pool size: ' + str(POOL_SIZE))
    # Restore state from checkpoint if provided


    s3_object_list = src.objects.all()
    num_objects_to_transfer = len(list(s3_object_list))
    logging.info(str(num_objects_to_transfer) + ' objects found to transfer')

    # Go through list once for folders; tqdm
    # Go through list second time for files (multiprocessing)
    with Manager() as manager:
        entities = manager.dict()

        if checkpoint is None:
            checkpoint = '.checkpoint.csv'

        _setup_checkpoint(checkpoint, entities)

        logging.info('Creating folders')
        file_paths = []
        for s3_object in progress(s3_object_list, desc='Create folders', unit='key'):
            if s3_object.key.endswith("/"):
                # s3_object is a Folder
                # e.g. s3_object.key --> 'Folder1/Folder2/Folder3/'

                folder_path = s3_object.key
                logging.info('Listed S3 folder: ' + folder_path)

                find_or_create_folder(entities, folder_path, project, session, checkpoint)
            else:
                file_paths.append(s3_object.key)

        logging.info('Folders created')

        logging.info('Transferring files')

        session_token = session.token

        # map_args = _prepare_args(file_paths, entities, project, session_token, source_s3_bucket, destination_s3_bucket,
        #                          aws_access_key_id, aws_secret_access_key, copy_file_fn, checkpoint)

        with Pool(POOL_SIZE) as p:
            partial_find_or_create_file = functools.partial(find_or_create_file, entities, project, session_token, source_s3_bucket, destination_s3_bucket,
                                                            aws_access_key_id, aws_secret_access_key, copy_file_fn, checkpoint)

            for r in progress(p.imap_unordered(partial_find_or_create_file, file_paths), desc='Copy objects', total=len(file_paths), unit='file'):
                pass
            # p.starmap(find_or_create_file, map_args, CHUNK_SIZE)

        return dict(entities)


def _prepare_args(file_paths, entities, project, session_token, source_s3_bucket, destination_s3_bucket,
                  aws_access_key_id, aws_secret_access_key, copy_file_fn, checkpoint):
    return ([entities, file_path, project, session_token, source_s3_bucket, destination_s3_bucket,
             aws_access_key_id, aws_secret_access_key, copy_file_fn, checkpoint] for (i,file_path) in enumerate(file_paths))


def _setup_checkpoint(checkpoint, entities):
    if os.path.isfile(checkpoint):
        with open(checkpoint, 'r') as f:
            reader = csv.DictReader(f)
            for row in reader:
                entities[row[PATH]] = row[ID]
            logging.info('Read {} entities from checkpoint file'.format(len(entities)))
    else:
        with open(checkpoint, 'w') as f:
            writer = csv.DictWriter(f, fieldnames=FIELDNAMES)
            writer.writeheader()


def find_or_create_file(entities, project, session_token, source_s3_bucket, destination_s3_bucket,
                        aws_access_key_id, aws_secret_access_key, copy_file_fn, checkpoint, file_path):
    with open(checkpoint, 'a') as checkpoint_file:
        checkpoint_writer = csv.DictWriter(checkpoint_file, fieldnames=FIELDNAMES)
        session = _make_session(session_token)

        if file_path not in entities:
            logging.info('Creating file: ' + file_path)

            path_list = file_path.split('/')

            # e.g file_name --> 'test.png'
            file_name = path_list[-1]

            if len(path_list) > 1:
                parent_folder_path = '/'.join(path_list[:-1]) + '/'
                logging.info('File parent folder is: ' + parent_folder_path)
                parent = find_or_create_folder(entities, parent_folder_path, project, session, checkpoint)
            else:
                logging.info('File parent is project')
                parent = project

            file = copy_file_fn(session, parent=parent, file_key=file_path, file_name=file_name,
                                source_bucket=source_s3_bucket, destination_bucket=destination_s3_bucket,
                                aws_access_key_id=aws_access_key_id,
                                aws_secret_access_key=aws_secret_access_key)

            entities[file_path] = file['_id']
            logging.info('Recording new file to checkpoing file: ' + file_path)
            checkpoint_writer.writerow({PATH: file_path, ID: file['_id']})
        else:
            logging.info('No need to create file: ' + file_path + '. Found in already created files list.')


def _make_session(session_token):
    session = Session(session_token)
    return session


def find_or_create_folder(entities, folder_path, project, session, checkpoint):
    with open(checkpoint, 'a') as checkpoint_file:
        checkpoint_writer = csv.DictWriter(checkpoint_file, fieldnames=FIELDNAMES)
        folder_list = folder_path.split('/')[:-1]  # Drop trailing

        if folder_path not in entities:

            logging.info('Creating folder: ' + folder_path)

            # e.g. current_folder --> 'Folder3'
            current_folder = folder_list[-1]
            parent_folder_path = '/'.join(folder_list[:-1]) + '/'

            # if nested folder
            if len(folder_list) > 1:
                logging.info('Folder parent_folder is: ' + parent_folder_path)
                parent = find_or_create_folder(entities, parent_folder_path, project, session, checkpoint)
            else:
                logging.info('Folder parent_folder is: project')
                parent = project

            found_or_created_folder = core.create_folder(session, parent, current_folder,
                                                         attributes={'original_s3_path': folder_path})

            entities[folder_path] = found_or_created_folder['_id']

            logging.info('Recording new folder to checkpoint file: ' + folder_path)
            checkpoint_writer.writerow({PATH: folder_path, ID: found_or_created_folder['_id']})
        else:
            logging.info(folder_path + ' already in entities')
            found_or_created_folder = entities[folder_path]

        return found_or_created_folder


def copy_file(session, parent=None, file_key=None, file_name=None, source_bucket=None,
              destination_bucket=None, aws_access_key_id=None, aws_secret_access_key=None):
    """
    Creates an Ovation 'File' and 'Revision' record.  File is transferred from
    source_bucket to destination_bucket and then the Revision` version is set to S3 version_id
    of the object in the destination_bucket.
    :param session: ovation.connection.Session
    :param parent: Project or Folder (entity dict or ID)
    :param file_key: S3 Key of file to copy from source bucket
    :param file_name: Name of file
    :param source_bucket: the S3 bucket to copy file from
    :param destination_bucket: the destination_bucket to copy file to
    :param aws_access_key_id: id for key that must have read access to source and write access to destination
    :param aws_secret_access_key: AWS key
    :return: new `Revision` entity dicitonary
    """
    content_type = upload.guess_content_type(file_name)

    # create file record
    new_file = core.create_file(session, parent, file_name,
                                attributes={'original_s3_path': file_key})

    # create revision record
    r = session.post(new_file.links.self,
                     data={'entities': [{'type': 'Revision',
                                         'attributes': {'name': file_name,
                                                        'content_type': content_type}}]})

    # get key for where to copy existing file from create revision resonse
    revision = r['entities'][0]
    destination_s3_key = r['aws'][0]['aws']['key']

    # Copy file from source bucket to destination location
    # Need to use key that has 'list' and 'get_object' privleges on source bucket as well as 'write' privleges
    # on destination bucket (can't use the temp aws key from the create revision response since this does
    # not have read access to the source bucket)
    aws_session = boto3.Session(aws_access_key_id=aws_access_key_id,
                                aws_secret_access_key=aws_secret_access_key)

    s3 = aws_session.resource('s3')
    destination_file = s3.Object(destination_bucket, destination_s3_key)
    copy_source = "{0}/{1}".format(source_bucket, file_key)
    aws_response = destination_file.copy_from(CopySource=copy_source)

    # get version_id from AWS copy response and update revision record with aws version_id
    # revision['attributes']['version'] = aws_response['VersionId']
    revision_response = session.put(revision['links']['upload-complete'], entity=None)

    return revision_response


def main():
    parser = argparse.ArgumentParser(description='Transfer files from S3 to Ovation')
    parser.add_argument('-u', '--user', help='Ovation user email')
    parser.add_argument('-p', '--password', help='Ovation password')
    parser.add_argument('-t', '--token', help='Ovation user token')
    parser.add_argument('parent', help='Project or Folder UUID that will receive transferred files')
    parser.add_argument('source_bucket', help='Source S3 Bucket')
    parser.add_argument('aws_access_key_id', help='AWS Access Key Id')
    parser.add_argument('aws_secret_access_key', help='AWS Secret Key')

    args = parser.parse_args()

    token = args.token
    if token is None:
        user = args.user
        if user is None:
            user = input('Email: ')

        session = connect(user, password=args.password)

    else:
        session = Session(token)

    aws_secret_access_key = args.aws_secret_access_key

    logging.basicConfig(level=logging.INFO, filename='ovation_transfer.log')

    copy_bucket_contents(session,
                         project=args.parent,
                         aws_access_key_id=args.aws_access_key_id,
                         aws_secret_access_key=aws_secret_access_key,
                         source_s3_bucket=args.source_bucket,
                         destination_s3_bucket='users.ovation.io',
                         checkpoint='.transfer_checkpoint.csv', copy_file_fn=copy_file)


if __name__ == '__main__':
    main()



#!/usr/bin/env python

import os

def is_conda():
    return 'CONDA_BUILD' in os.environ

if is_conda():
    from distutils.core import setup
else:
    from setuptools import setup

DESCRIPTION =  """
Ovation is a Scientific Data Layer-as-a-Service, engineered specifically for scientists. Ovation helps researchers manage files and metadata, link raw data with analysis and share all of this with colleagues and collaborators.

The Ovation Python API wraps the Ovation REST API.
"""


with open(os.path.join('ovation', '__init__.py')) as fd:
    versionline = [x for x in fd.readlines() if x.startswith('__version__')]
    version = versionline[0].split("'")[-2]


args = dict(name='ovation',
            version=version,
            description='Ovation Python API',
            author='Physion LLC',
            author_email='info@ovation.io',
            url='https://ovation.io',
            long_description=DESCRIPTION,
            packages=['ovation'],
            classifiers=[
          "Development Status :: 4 - Beta",
          "Intended Audience :: Science/Research",
          "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
      ])


if not is_conda():
    args.update(zip_safe=False,
                setup_requires=['nose>=1.3.7', 'coverage>=4.0.3'],
                install_requires=["requests >= 2.9.1",
                                  "six >= 1.10.0",
                                  "boto3 >= 1.3.0",
                                  'six >= 1.10.0',
                                  'tqdm >= 4.7.0',
                                  'retrying >= 1.3.3',
                                  'texttable >= 0.8.4'
                                  ],
                test_suite='nose.collector')

setup(**args)





import copy
from unittest.mock import Mock, sentinel, patch

from nose.tools import istest, assert_equal

import ovation.session
import ovation.upload as revisions

@istest
@patch('boto3.Session')
def should_create_revision(boto_session):
    file = {'type': 'File',
            'links': {'self': sentinel.self}}
    path = '/local/path/file.txt'
    rev = {'_id': 1,
           'type': 'Revision',
           'attributes': {'url': sentinel.url},
           'links': {'upload-complete': sentinel.upload_complete}}

    s = Mock(spec=ovation.session.Session)

    def entity_path(type='', id=''):
        return "/api/v1/{}/{}".format(type, id)

    s.entity_path.side_effect = entity_path
    aws_session = Mock()
    s3 = Mock()
    boto_session.return_value = aws_session
    aws_session.resource = Mock(return_value=s3)

    file_obj = Mock()
    s3.Object = Mock(return_value=file_obj)
    file_obj.upload_file = Mock()
    file_obj.version_id = sentinel.version

    s.post = Mock(return_value={'entities': [rev],
                                'aws': [{'aws': dict(access_key_id=sentinel.access_key,
                                                     secret_access_key=sentinel.secret_key,
                                                     session_token=sentinel.session_token,
                                                     bucket=sentinel.bucket,
                                                     key=sentinel.key)}]})

    s.put = Mock(return_value=sentinel.result)

    # Act
    result = revisions.upload_revision(s, file, path)

    # Assert
    boto_session.assert_called_with(aws_access_key_id=sentinel.access_key,
                                    aws_secret_access_key=sentinel.secret_key,
                                    aws_session_token=sentinel.session_token)
    s3.Object.assert_called_with(sentinel.bucket, sentinel.key)
    file_obj.upload_file.assert_called_with(path, ExtraArgs={'ContentType': 'text/plain',
                                                             'ServerSideEncryption': 'AES256'})

    s.put.assert_called_with(sentinel.upload_complete, entity=None)

    assert_equal(result, sentinel.result)




